# frozen_string_literal: true

require 'applitools/selenium/stitch_modes'
require 'applitools/selenium/browsers_info'

module Applitools
  module Selenium
    class Configuration < Applitools::EyesBaseConfiguration
      DEFAULT_CONFIG = proc do
        {
          # force_full_page_screenshot: false,
          wait_before_screenshots: 0.1,
          # stitch_mode: Applitools::Selenium::StitchModes::CSS,
          hide_scrollbars: true,
          hide_caret: false,
          browsers_info: Applitools::Selenium::BrowsersInfo.new,
          rendering_grid_force_put: (ENV['APPLITOOLS_RENDERING_GRID_FORCE_PUT'] || 'false').casecmp('true') == 0,
          visual_grid_options: {}
        }
      end
      class << self
        def default_config
          super.merge DEFAULT_CONFIG.call
        end
      end

      boolean_field :force_full_page_screenshot
      int_field :wait_before_screenshots
      int_field :wait_before_capture
      enum_field :stitch_mode, Applitools::Selenium::StitchModes.enum_values
      boolean_field :hide_scrollbars
      boolean_field :hide_caret
      boolean_field :send_dom

      object_field :browsers_info, Applitools::Selenium::BrowsersInfo
      int_field :concurrent_sessions
      boolean_field :rendering_grid_force_put
      object_field :visual_grid_options, Hash, true

      def custom_setter_for_visual_grid_options(value)
        return {} if value.nil?
        value
      end

      def add_multi_device_target(*args)
        ios_devices = []

        if args.length == 1 && args[0].is_a?(Hash)
          # Single hash passed directly
          ios_devices << create_ios_device_info(args[0])
        elsif args.any? { |arg| arg.is_a?(Hash) }
          # One of the arguments is a hash with device configuration
          hash_arg = args.find { |arg| arg.is_a?(Hash) }
          ios_devices << create_ios_device_info(hash_arg)
        else
          # Multiple device names as symbols or strings
          args.each do |device_name|
            ios_devices << create_ios_device_info(device_name: device_name)
          end
        end

        # Add each iOS device to the configuration
        add_browsers(*ios_devices)
        # Return the configuration object for method chaining
        self
      end

      def add_browser(*args)
        case args.size
        when 0
          browser = Applitools::Selenium::DesktopBrowserInfo.new
        when 1
          b = args[0]
          raise(
            Applitools::EyesIllegalArgument,
            'Expected :browser to be an IRenderBrowserInfo instance!'
          ) unless b.is_a? IRenderBrowserInfo
          browser = b
        when 3
          browser = Applitools::Selenium::DesktopBrowserInfo.new.tap do |bi|
            bi.viewport_size = Applitools::RectangleSize.new(args[0], args[1])
            bi.browser_type = args[2]
          end
        end
        yield(Applitools::Selenium::RenderBrowserInfoFluent.new(browser)) if block_given?
        browsers_info.add browser
        # self.viewport_size = browser.viewport_size unless viewport_size
        self
      end

      def add_browsers(*browsers)
        browsers = case browsers.first
                     when Applitools::Selenium::IRenderBrowserInfo
                       browsers
                     when Array
                       browsers.first
                   end
        browsers.each do |browser|
          add_browser(browser)
        end
        self
      end

      def add_device_emulation(device_name, orientation = Orientations::PORTRAIT)
        Applitools::ArgumentGuard.not_nil device_name, 'device_name'
        raise Applitools::EyesIllegalArgument, 'Wrong device name!' unless Devices.enum_values.include? device_name
        emu = Applitools::Selenium::ChromeEmulationInfo.new(device_name, orientation)
        add_browser emu
      end

      def add_mobile_device(mobile_device_info)
        add_mobile_devices(mobile_device_info)
      end

      def add_mobile_devices(mobile_device_infos)
        add_browsers(mobile_device_infos)
      end

      # Move viewport_size above the private methods
      def viewport_size
        user_defined_vp = super
        user_defined_vp = nil if user_defined_vp.respond_to?(:square) && user_defined_vp.square == 0
        return user_defined_vp if user_defined_vp
        # from_browsers_info = browsers_info.select { |bi| bi.viewport_size.square > 0 }.first
        # return from_browsers_info.viewport_size if from_browsers_info
        nil
      end

      private

      def create_ios_device_info(options)
        if options.is_a?(Symbol)
          # If a symbol is passed, treat it as a device name
          Applitools::Selenium::IosDeviceInfo.new(device_name: options)
        else
          # Otherwise, pass the options hash to IosDeviceInfo
          Applitools::Selenium::IosDeviceInfo.new(options)
        end
      end

      def handle_device_target(target)
        if target.is_a?(Applitools::Selenium::IosDeviceTarget)
          ios_device_info = Applitools::Selenium::IosDeviceInfo.new(
            device_name: target.device_name,
            screen_orientation: target.screen_orientation,
            ios_version: target.version
          )
          add_browser(ios_device_info)
        elsif target.is_a?(Applitools::Selenium::DeviceTarget)
          device_name = target.device_name
          orientation = target.screen_orientation || Orientations::PORTRAIT
          add_device_emulation(device_name, orientation)
        else
          raise Applitools::EyesIllegalArgument,
                'Expected arguments to be device names, hashes with configuration, or DeviceTarget instances'
        end
      end

    end
  end
end
